<?php // phpcs:ignore

namespace WP2FA\Admin;

use \WP2FA\WP2FA as WP2FA;
use \WP2FA\Authenticator\Authentication as Authentication;
use \WP2FA\Admin\SettingsPage as SettingsPage;

/**
 * UserProfile - Class for handling user things such as profile settings and admin list views.
 */
class UserRegistered {

	/**
	 * Classs constructor
	 */
	public function __construct() {

	}

	/**
	 * Apply 2FA Grace period
	 *
	 * @param  int $user_id User id.
	 */
	public function apply_2fa_grace_period( $user_id ) {
		// Get user object.
		$user = get_user_by( 'id', $user_id );
		// Check if this user is actually eligible.
		$is_needed        = Authentication::is_user_eligible_for_2fa( $user->ID );
		$is_user_excluded = WP2FA::is_user_excluded( $user->ID );
		// If they are, add grace_period.
		if ( $is_needed && ! $is_user_excluded ) {
			// Grab grace period.
			$create_a_string = WP2FA::get_wp2fa_setting( 'grace-period' ) . ' ' . WP2FA::get_wp2fa_setting( 'grace-period-denominator' );
			// Turn that string into a time.
			$grace_expiry = strtotime( $create_a_string );
			update_user_meta( $user->ID, 'wp_2fa_grace_period_expiry', $grace_expiry );
			SettingsPage::send_2fa_enforced_email( $user->ID );
		}
	}

	public function check_user_upon_role_change( $user_id, $role, $old_roles ) {
		$this->apply_2fa_grace_period( $user_id );
	}
}
